"""
This decorator finds fields of an object, assigns them to slots,
and disables modification for fields which begin with `_`. This
makes these fields impossible to modify outside initialization.

TODO: allow custom init, even if a little bit clunky
"""

from __future__ import annotations
from typing import TypeVar, List
from inspect import get_annotations

T = TypeVar('T')


class WrongInit(Exception):
    pass


class SafeClass:
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        pass


def safeclass(cls: type) -> type:
    var_names: List[str] = list()

    cls_annotations = get_annotations(cls)
    for name, annotation_val in cls_annotations.items():
        if isinstance(annotation_val, str):
            try:
                cls_annotations[name] = eval(annotation_val)
            except NameError:
                # we do not have enough information, we have
                # to be permissive
                cls_annotations[name] = object

    for name in cls.__annotations__:
        var_names.append(name)

    def _init(self, *args, **kwargs):
        for i, value in enumerate(args):
            if i >= len(var_names):
                raise WrongInit("Too many arguments")
            name = var_names[i]
            if not isinstance(value, cls_annotations[name]):
                raise WrongInit("Argument types do not match")
            super(self.__class__, self).__setattr__(name, value)

        for name in var_names[len(args):]:
            if name in kwargs:
                value = kwargs[name]
            elif name in vars(cls):
                value = cls.name
            else:
                raise WrongInit(f"Missing an argument {name}")

            if not isinstance(value, cls_annotations[name]):
                raise WrongInit("Argument types do not match")

            super(self.__class__, self).__setattr__(name, value)

    def _setattr(self, name: str, val: 'T'):
        if name.startswith('_'):
            raise AttributeError("Cannot modify an immutable attribute")
        super(self.__class__, self).__setattr__(name, val)

    cls_dict = dict(cls.__dict__)
    cls_dict['__slots__'] = var_names
    cls_dict['__init__'] = _init
    cls_dict['__setattr__'] = _setattr

    return type(cls.__name__, cls.__bases__, cls_dict)


def test() -> None:
    @safeclass
    class Person(SafeClass):
        _name: str
        age: int

    @safeclass
    class Point(SafeClass):
        x: int
        y: int

        def __str__(self) -> str:
            return f"Point(x = {self.x}, y = {self.y})"

    @safeclass
    class Line(SafeClass):
        a: Point
        b: Point

        def __str__(self) -> str:
            return f"Line(a={self.a}, b={self.b})"

    line = Line(Point(0, 0), Point(2, 3))
    # print(line)

    p = Person("Joe", 30)

    # print(p.__dict__)
    # print(vars(Person))

    p.age += 1

    try:
        p.what = 5
        print(p.__dict__, p.__class__, p.__class__.__dict__)
        assert False
    except AttributeError:
        print("all is well")

    try:
        p._name = "Mama"
        assert False
    except AttributeError:
        print("all is well")


if __name__ == '__main__':
    test()
